# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import select
from contextlib import closing

import sys
import os
import re
import random
import subprocess
import socket
import errno
from functools import wraps
from collections import namedtuple
from contextlib import contextmanager
from subprocess import call
if sys.version_info[:2] < (3, 2):
    from contextlib import nested

import mock
import pytest

import pyudev


if sys.version_info[:2] >= (3, 2):
    @contextmanager
    def nested(*managers):
        """
        Copy of contextlib.nested from python 2.7 with slight adaptions to
        Python 3.2 to use nested() on python 3.2 and thus retain backwards
        compatibility with Python 2.6 (with doesn't support the new nesting
        syntax), while being able to run the test suite on Python 3.2 as
        well (which doesn't provide contextlib.nested anymore).
        """
        exits = []
        vars = []
        exc = None
        try:
            for mgr in managers:
                exit = mgr.__exit__
                enter = mgr.__enter__
                vars.append(enter())
                exits.append(exit)
            yield vars
        except BaseException as e:
            exc = e
        finally:
            while exits:
                exit = exits.pop()
                try:
                    if exc:
                        args = (type(exc), exc, exc.__traceback__)
                    else:
                        args = (None, None, None)
                    if exit(*args):
                        exc = None
                except BaseException as e:
                    exc = e
            if exc:
                raise exc


class FakeMonitor(object):
    """
    A dummy pyudev.Monitor class, which allows clients to trigger arbitrary
    events, emitting clearly defined device objects.
    """

    def __init__(self, device_to_emit):
        self.client, self.server = socket.socketpair(
            socket.AF_UNIX, socket.SOCK_DGRAM)
        if sys.version_info[0] >= 3:
            # in python 3 sockets returned by socketpair() lack the
            # ".makefile()" method, which is required by this class.  Work
            # around this limitation by wrapping these sockets in real
            # socket objects.
            def _wrap_socket(sock):
                wrapped = socket.socket(sock.family, sock.type,
                                        fileno=os.dup(sock.fileno()))
                sock.close()
                return wrapped
            self.client, self.server = (_wrap_socket(self.client),
                                        _wrap_socket(self.server))
        self.device_to_emit = device_to_emit

    def trigger_action(self, action):
        """
        Trigger the given ``action`` on clients of this monitor.
        """
        with self.server.makefile('w') as stream:
            stream.write(action)
            stream.write('\n')
            stream.flush()

    def fileno(self):
        return self.client.fileno()

    def enable_receiving(self):
        pass

    def filter_by(self, *args):
        pass

    start = enable_receiving

    def receive_device(self):
        with self.client.makefile('r') as stream:
            action = stream.readline().strip()
            return action, self.device_to_emit

    def close(self):
        """
        Close sockets acquired by this monitor.
        """
        try:
            self.client.close()
        finally:
            self.server.close()


class UDevAdm(object):
    """
    Wrap ``udevadm`` utility.
    """

    CANDIDATES = ['/sbin/udevadm', 'udevadm']

    @classmethod
    def find(cls):
        for candidate in cls.CANDIDATES:
            try:
                return cls(candidate)
            except EnvironmentError as error:
                if error.errno != errno.ENOENT:
                    raise

    def __init__(self, udevadm):
        """
        Create a new ``udevadm`` wrapper for the given udevadm executable.

        ``udevadm`` is the path to udevadm as string.  If relative, ``udevadm`` is
        looked up in ``$PATH``.
        """
        self.udevadm = udevadm
        self.version = int(self._execute('--version'))

    def _execute(self, *args):
        command = [self.udevadm] + list(args)
        proc = subprocess.Popen(command, stdout=subprocess.PIPE)
        output = proc.communicate()[0].strip()
        if proc.returncode != 0:
            raise subprocess.CalledProcessError(proc.returncode, command)
        return output

    def read_database(self, properties_blacklist):
        database = self._execute('info', '--export-db').splitlines()
        devices = {}
        current_properties = None
        for line in database:
            line = line.strip().decode(sys.getfilesystemencoding())
            if not line:
                continue
            type, value = line.split(': ', 1)
            if type == 'P':
                current_properties = devices.setdefault(value, {})
            elif type == 'E':
                property, value = value.split('=', 1)
                if property in properties_blacklist:
                    continue
                current_properties[property] = value
        return devices

    def get_device_attributes(self, device_path, attributes_blacklist):
        output = self._execute('info', '--attribute-walk',
                               '--path', device_path)
        attribute_dump = output.splitlines()
        attributes = {}
        for line in attribute_dump:
            line = line.strip().decode(sys.getfilesystemencoding())
            if line.startswith('looking at parent device'):
                # we don't continue with attributes of parent devices, we only
                # want the attributes of the given device
                break
            if line.startswith('ATTR'):
                name, value = line.split('==', 1)
                # remove quotation marks from attribute value
                value = value[1:-1]
                # remove prefix from attribute name
                name = re.search('{(.*)}', name).group(1)
                if name in attributes_blacklist:
                    continue
                attributes[name] = value
        return attributes

    def query_device(self, device_path, query_type):
        if query_type not in ('symlink', 'name'):
            raise ValueError(query_type)
        try:
            output = self._execute('info', '--root', '--path', device_path,
                                   '--query', query_type)
        except subprocess.CalledProcessError:
            return None
        else:
            query_result = output.decode(sys.getfilesystemencoding())
            if query_type == 'symlink':
                return query_result.split()
            else:
                return query_result


def get_device_sample(config):
    if config.getvalue('device'):
        return [config.getvalue('device')]
    if config.getvalue('all_devices'):
        return config.udev_database
    else:
        device_sample_size = config.getvalue('device_sample_size')
        actual_size = min(device_sample_size, len(config.udev_database))
        return random.sample(list(config.udev_database), actual_size)


def assert_env_error(error, errno, filename=None):
    __tracebackhide__ = True
    assert error.errno == errno
    assert error.strerror == os.strerror(errno)
    assert error.filename == filename


@contextmanager
def patch_libudev(funcname):
    with mock.patch('pyudev._libudev.libudev.{0}'.format(funcname)) as func:
        yield func


class Node(namedtuple('Node', 'name next')):
    """
    Node in a :class:`LinkedList`.
    """
    @property
    def value(self):
        return 'value of {0}'.format(self.name)


class LinkedList(object):
    """
    Linked list class to mock libudev list functions.
    """

    @classmethod
    def from_sequence(cls, items):
        next_node = None
        for item in reversed(items):
            node = Node(item, next_node)
            next_node = node
        return cls(next_node)

    def __init__(self, first):
        self.first = first


@contextmanager
def patch_libudev_list(items, list_func):
    linked_list = LinkedList.from_sequence(items)
    get_next = 'udev_list_entry_get_next'
    get_name = 'udev_list_entry_get_name'
    get_value = 'udev_list_entry_get_value'
    with pytest.nested(pytest.patch_libudev(get_next),
                       pytest.patch_libudev(get_name),
                       pytest.patch_libudev(get_value),
                       pytest.patch_libudev(list_func)) as (get_next, get_name,
                                                            get_value,
                                                            list_func):
        list_func.return_value = linked_list.first
        get_name.side_effect = lambda e: e.name
        get_value.side_effect = lambda e: e.value
        get_next.side_effect = lambda e: e.next
        yield list_func


def _need_privileges(func):
    @wraps(func)
    def wrapped(*args):
        if not pytest.config.getvalue('allow_privileges'):
            pytest.skip('privileges disabled')
        return func(*args)
    return wrapped


@_need_privileges
def load_dummy():
    """
    Load the ``dummy`` module or raise :exc:`ValueError` if sudo are not
    allowed.
    """
    call(['sudo', 'modprobe', 'dummy'])


@_need_privileges
def unload_dummy():
    """
    Unload the ``dummy`` module or raise :exc:`ValueError` if sudo are not
    allowed.
    """
    call(['sudo', 'modprobe', '-r', 'dummy'])


def is_unicode_string(value):
    """
    Return ``True``, if ``value`` is of a real unicode string type
    (``unicode`` in python 2, ``str`` in python 3), ``False`` otherwise.
    """
    if sys.version_info[0] >= 3:
        unicode_type = str
    else:
        unicode_type = unicode
    return isinstance(value, unicode_type)


def need_udev_version(version_spec):
    """
    Decorator to check the udev version before a test is run.

    ``version_spec`` is a string containing a comparison operator and the right
    hand sight operand of this comparison, like ``'>= 165'``.  The left hand
    side is automatically provided by the current udev version.

    If the udev version matches the spec, the test is run.  Otherwise the test
    is skipped.
    """
    expr_template = 'not ({0} {1})'
    actual_version = pyudev.udev_version()
    expr = expr_template.format(actual_version, version_spec)
    reason = 'udev version mismatch: {0} required, {1} found'.format(
        version_spec, actual_version)
    return pytest.mark.skipif(str(expr), reason=reason)


def pytest_namespace():
    return dict((func.__name__, func) for func in
                (get_device_sample, patch_libudev, load_dummy, nested,
                 unload_dummy, is_unicode_string, need_udev_version,
                 patch_libudev_list, assert_env_error))


def pytest_addoption(parser):
    parser.addoption('--all-devices', action='store_true',
                     help='Run device tests against *all* devices in the '
                     'database.  By default, only a random sample will be '
                     'checked.', default=False)
    parser.addoption('--device', metavar='DEVICE',
                     help='Run the device tests only against the given '
                     'DEVICE', default=None)
    parser.addoption('--device-sample-size', type='int', metavar='N',
                     help='Use a random sample of N elements (default: 10)',
                     default=10)
    parser.addoption('--allow-privileges', action='store_true',
                     help='Permit execution of tests, which require '
                     'root privileges.  This affects all monitor tests, '
                     'that need to load or unload modules to trigger udev '
                     'events.  "sudo" will be used to execute privileged '
                     'code, be sure to have proper privileges before '
                     'enabling this option!', default=False)


def pytest_configure(config):
    # these are volatile, frequently changing properties and attributes,
    # which lead to bogus failures during tests, and therefore they are
    # masked and shall be ignored for test runs.
    config.properties_blacklist = frozenset(
        ['POWER_SUPPLY_CURRENT_NOW', 'POWER_SUPPLY_VOLTAGE_NOW',
         'POWER_SUPPLY_CHARGE_NOW'])
    config.attributes_blacklist = frozenset(
        ['power_on_acct', 'temp1_input', 'charge_now', 'current_now',
         'urbnum'])
    config.udevadm = UDevAdm.find()
    config.udev_database = config.udevadm.read_database(config.properties_blacklist)
    config.udev_version = config.udevadm.version


def pytest_funcarg__database(request):
    """
    The complete udev database parsed from the output of ``udevadm info
    --export-db``.

    Return a dictionary, mapping the devpath of a device *without* sysfs
    mountpoint to a dictionary of properties of the device.
    """
    return request.config.udev_database


def pytest_funcarg__context(request):
    """
    Return a useable :class:`pyudev.Context` object.  The context is cached
    with session scope.
    """
    return request.cached_setup(setup=pyudev.Context, scope='session')


def pytest_funcarg__device_path(request):
    """
    Return a device path as string.

    The device path must be available as ``request.param``.
    """
    return request.param


def pytest_funcarg__all_properties(request):
    """
    Get all properties from the exported database (as returned by the
    ``database`` funcarg) of the device pointed to by the ``device_path``
    funcarg.
    """
    device_path = request.getfuncargvalue('device_path')
    return dict(request.getfuncargvalue('database')[device_path])


def pytest_funcarg__properties(request):
    """
    Same as the ``all_properties`` funcarg, but with the special ``DEVNAME``
    property removed.
    """
    properties = request.getfuncargvalue('all_properties')
    properties.pop('DEVNAME', None)
    return properties


def pytest_funcarg__tags(request):
    """
    Return the tags of the device pointed to by the ``device_path`` funcarg.
    """
    properties = request.getfuncargvalue('properties')
    tags = properties.get('TAGS', '')
    return [t for t in tags.split(':') if t]


def pytest_funcarg__attributes(request):
    """
    Return a dictionary of all attributes for the device pointed to by the
    ``device_path`` funcarg.
    """
    device_path = request.getfuncargvalue('device_path')
    return request.config.udevadm.get_device_attributes(
        device_path, request.config.attributes_blacklist)


def pytest_funcarg__device_node(request):
    """
    Return the name of the device node for the device pointed to by the
    ``device_path`` funcarg.
    """
    device_path = request.getfuncargvalue('device_path')
    return request.config.udevadm.query_device(device_path, 'name')


def pytest_funcarg__device_number(request):
    """
    Return the device number of the device pointed to by the ``device_path``
    funcarg.
    """
    device_node = request.getfuncargvalue('device_node')
    if device_node:
        return os.stat(device_node).st_rdev
    else:
        return 0


def pytest_funcarg__device_links(request):
    """
    Return a list of symlink to the device node (``device_node`` funcarg)
    for the device pointed to by the ``device_path`` funcarg.
    """
    device_path = request.getfuncargvalue('device_path')
    return request.config.udevadm.query_device(device_path, 'symlink')


def pytest_funcarg__sys_path(request):
    """
    Return the sys_path including the sysfs mountpoint for the device path
    returned by the ``device_path`` funcarg.
    """
    context = request.getfuncargvalue('context')
    device_path = request.getfuncargvalue('device_path')
    return context.sys_path + device_path


def pytest_funcarg__device(request):
    """
    Create and return a :class:`pyudev.Device` object for the sys_path
    returned by the ``sys_path`` funcarg, and the context from the
    ``context`` funcarg.
    """
    sys_path = request.getfuncargvalue('sys_path')
    context = request.getfuncargvalue('context')
    return pyudev.Device.from_sys_path(context, sys_path)


def pytest_funcarg__platform_device(request):
    """
    Return the platform device at ``/sys/devices/platform``.  This device
    object exists on all systems, and is therefore suited to for testing
    purposes.
    """
    context = request.getfuncargvalue('context')
    return pyudev.Device.from_sys_path(context, '/sys/devices/platform')


def pytest_funcarg__socket_path(request):
    """
    Return a socket path for :meth:`pyudev.Monitor.from_socket`.  The path
    is unique for each test.
    """
    tmpdir = request.getfuncargvalue('tmpdir')
    return tmpdir.join('monitor-socket')


def pytest_funcarg__monitor(request):
    """
    Return a netlink monitor for udev source.
    """
    return pyudev.Monitor.from_netlink(request.getfuncargvalue('context'))


def pytest_funcarg__fake_monitor(request):
    """
    Return a FakeMonitor, which emits the platform device as returned by
    the ``platform_device`` funcarg on all triggered actions.
    """
    return FakeMonitor(request.getfuncargvalue('platform_device'))



# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

from functools import partial

import pytest
import mock

from pyudev import Enumerator


def pytest_generate_tests(metafunc):
    if 'device' in metafunc.funcargnames:
        devices = pytest.get_device_sample(metafunc.config)
        for device_path in devices:
            metafunc.addcall(id=device_path, param=device_path)


def pytest_funcarg__enumerator(request):
    context = request.getfuncargvalue('context')
    return context.list_devices()


class TestEnumerator(object):

    def test_match_subsystem(self, context):
        devices = context.list_devices().match_subsystem('input')
        assert all(d.subsystem == 'input' for d in devices)

    def test_match_subsystem_nomatch(self, context):
        devices = context.list_devices().match_subsystem('input', nomatch=True)
        assert all(d.subsystem != 'input' for d in devices)

    def test_match_subsystem_nomatch_unfulfillable(self, context):
        devices = context.list_devices()
        devices.match_subsystem('input')
        devices.match_subsystem('input', nomatch=True)
        assert not list(devices)

    def test_match_sys_name(self, context):
        devices = context.list_devices().match_sys_name('sda')
        assert all(d.sys_name == 'sda' for d in devices)

    def test_match_property_string(self, context):
        devices = list(context.list_devices().match_property('DRIVER', 'usb'))
        assert all(d['DRIVER'] == 'usb' for d in devices)
        assert all(d.driver == 'usb' for d in devices)

    def test_match_property_int(self, context):
        devices = list(context.list_devices().match_property(
            'ID_INPUT_KEY', 1))
        assert all(d['ID_INPUT_KEY'] == '1' for d in devices)
        assert all(d.asint('ID_INPUT_KEY') == 1 for d in devices)

    def test_match_property_bool(self, context):
        devices = list(context.list_devices().match_property(
            'ID_INPUT_KEY', True))
        assert all(d['ID_INPUT_KEY'] == '1' for d in devices)
        assert all(d.asbool('ID_INPUT_KEY') for d in devices)

    def test_match_attribute_nomatch(self, context):
        devices = context.list_devices().match_attribute(
            'driver', 'usb', nomatch=True)
        assert all(d.attributes.get('driver') != 'usb' for d in devices)

    def test_match_attribute_nomatch_unfulfillable(self, context):
        devices = context.list_devices()
        devices.match_attribute('driver', 'usb')
        devices.match_attribute('driver', 'usb', nomatch=True)
        assert not list(devices)

    def test_match_attribute_string(self, context):
        devices = list(context.list_devices().match_attribute('driver', 'usb'))
        assert all(d.attributes['driver'] == b'usb' for d in devices)

    def test_match_attribute_int(self, context):
        # busnum gives us the number of a USB bus.  And any decent system
        # likely has two or more usb buses, so this should work on more or less
        # any system.  I didn't find any other attribute that is likely to be
        # present on a wide range of system, so this is probably as general as
        # possible.  Still it may fail because the attribute isn't present on
        # any device at all on the system running the test
        devices = list(context.list_devices().match_attribute('busnum', 2))
        assert all(d.attributes['busnum'] == b'2' for d in devices)
        assert all(d.attributes.asint('busnum') == 2 for d in devices)

    def test_match_attribute_bool(self, context):
        # ro tells us whether a volumne is mounted read-only or not.  And any
        # developers system should have at least one readable volume, thus this
        # test should work on all systems these tests are ever run on
        devices = list(context.list_devices().match_attribute('ro', False))
        assert all(d.attributes['ro'] == b'0' for d in devices)
        assert all(not d.attributes.asbool('ro') for d in devices)

    @pytest.need_udev_version('>= 154')
    def test_match_tag_mock(self, context):
        add_match_tag = 'udev_enumerate_add_match_tag'
        enumerator = context.list_devices()
        with pytest.patch_libudev(add_match_tag) as add_match_tag:
            retval = enumerator.match_tag('spam')
            assert retval is enumerator
            add_match_tag.assert_called_with(enumerator, b'spam')
            args, _ = add_match_tag.call_args
            assert isinstance(args[1], bytes)

    @pytest.need_udev_version('>= 154')
    def test_match_tag(self, context):
        devices = list(context.list_devices().match_tag('seat'))
        assert all('seat' in d.tags for d in devices)

    @pytest.need_udev_version('>= 172')
    def test_match_parent(self, context, device):
        parent = device.parent
        if parent is None:
            pytest.skip('Device {0!r} has no parent'.format(device))
        else:
            children = list(context.list_devices().match_parent(parent))
            assert device in children
            assert parent in children

    @pytest.need_udev_version('>= 165')
    def test_match_is_initialized(self, context):
        match_is_initialized = 'udev_enumerate_add_match_is_initialized'
        with pytest.patch_libudev(match_is_initialized) as match_is_initialized:
            context.list_devices().match_is_initialized()
            assert match_is_initialized.called

    def test_combined_matches_of_same_type(self, context):
        """
        Test for behaviour as observed in #1
        """
        properties = ('DEVTYPE', 'ID_TYPE')
        devices = context.list_devices()
        for property in properties:
            devices.match_property(property, 'disk')
        assert all(any(d.get(p) == 'disk' for p in properties) for d in devices)

    def test_combined_matches_of_different_types(self, context):
        properties = ('DEVTYPE', 'ID_TYPE')
        devices = context.list_devices().match_subsystem('input')
        for property in properties:
            devices.match_property(property, 'disk')
        devices = list(devices)
        assert not devices

    def test_match(self, context):
        devices = list(context.list_devices().match(
            subsystem='input', ID_INPUT_MOUSE=True, sys_name='mouse0'))
        assert all(d.subsystem == 'input' for d in devices)
        assert all(d.asbool('ID_INPUT_MOUSE') for d in devices)
        assert all(d.sys_name == 'mouse0' for d in devices)

    def test_match_passthrough_subsystem(self, enumerator):
        with mock.patch.object(enumerator, 'match_subsystem',
                               mocksignature=True) as match_subsystem:
            enumerator.match(subsystem=mock.sentinel.subsystem)
            match_subsystem.assert_called_with(mock.sentinel.subsystem, False)

    def test_match_passthrough_sys_name(self, enumerator):
        with mock.patch.object(enumerator, 'match_sys_name',
                               mocksignature=True) as match_sys_name:
            enumerator.match(sys_name=mock.sentinel.sys_name)
            match_sys_name.assert_called_with(mock.sentinel.sys_name)

    def test_match_passthrough_tag(self, enumerator):
        with mock.patch.object(enumerator, 'match_tag',
                               mocksignature=True) as match_tag:
            enumerator.match(tag=mock.sentinel.tag)
            match_tag.assert_called_with(mock.sentinel.tag)

    @pytest.need_udev_version('>= 172')
    def test_match_passthrough_parent(self, enumerator):
        with mock.patch.object(enumerator, 'match_parent',
                               mocksignature=True) as match_parent:
            enumerator.match(parent=mock.sentinel.parent)
            match_parent.assert_called_with(mock.sentinel.parent)

    def test_match_passthrough_property(self, enumerator):
        with mock.patch.object(enumerator, 'match_property',
                               mocksignature=True) as match_property:
            enumerator.match(eggs=mock.sentinel.eggs, spam=mock.sentinel.spam)
            assert match_property.call_count == 2
            posargs = [args for args, _ in match_property.call_args_list]
            assert ('spam', mock.sentinel.spam) in posargs
            assert ('eggs', mock.sentinel.eggs) in posargs


class TestContext(object):

    @pytest.mark.match
    def test_list_devices(self, context):
        devices = list(context.list_devices(subsystem='input', ID_INPUT_MOUSE=True,
                                            sys_name='mouse0'))
        assert all(d.subsystem == 'input' for d in devices)
        assert all(d.asbool('ID_INPUT_MOUSE') for d in devices)
        assert all(d.sys_name == 'mouse0' for d in devices)

    @pytest.mark.match
    def test_list_devices_passthrough(self, context):
        with mock.patch.object(Enumerator, 'match') as match:
            context.list_devices(subsystem=mock.sentinel.subsystem,
                                 sys_name=mock.sentinel.sys_name,
                                 tag=mock.sentinel.tag,
                                 parent=mock.sentinel.parent,
                                 prop1=mock.sentinel.prop1,
                                 prop2=mock.sentinel.prop2)
            match.assert_called_with(
                subsystem=mock.sentinel.subsystem,
                sys_name=mock.sentinel.sys_name,
                tag=mock.sentinel.tag,
                parent=mock.sentinel.parent,
                prop1=mock.sentinel.prop1,
                prop2=mock.sentinel.prop2)



# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os
import sys
import errno

import pytest

from pyudev import _util


@pytest.mark.conversion
def test_ensure_byte_string():
    assert isinstance(_util.ensure_byte_string('hello world'), bytes)
    assert _util.ensure_byte_string('hello world') == b'hello world'
    hello = b'hello world'
    assert _util.ensure_byte_string(hello) is hello


@pytest.mark.conversion
def test_ensure_byte_string_none():
    with pytest.raises(AttributeError):
        _util.ensure_byte_string(None)


@pytest.mark.conversion
def test_ensure_unicode_string():
    assert pytest.is_unicode_string(
        _util.ensure_unicode_string(b'hello world'))
    assert _util.ensure_unicode_string(b'hello world') == 'hello world'
    hello = 'hello world'
    assert _util.ensure_unicode_string(hello) is hello


@pytest.mark.conversion
def test_ensure_unicode_string_none():
    with pytest.raises(AttributeError):
        _util.ensure_unicode_string(None)


@pytest.mark.conversion
def test_property_value_to_bytes_string():
    hello = 'hello world'.encode(sys.getfilesystemencoding())
    assert _util.property_value_to_bytes(hello) is hello
    assert isinstance(_util.property_value_to_bytes('hello world'), bytes)
    assert _util.property_value_to_bytes('hello world') == hello


@pytest.mark.conversion
def test_property_value_to_bytes_int():
    assert _util.property_value_to_bytes(10000) == b'10000'
    assert isinstance(_util.property_value_to_bytes(10000), bytes)


@pytest.mark.conversion
def test_property_value_to_bytes_bool():
    assert _util.property_value_to_bytes(True) == b'1'
    assert isinstance(_util.property_value_to_bytes(True), bytes)
    assert _util.property_value_to_bytes(False) == b'0'
    assert isinstance(_util.property_value_to_bytes(False), bytes)


@pytest.mark.conversion
def test_string_to_bool_true():
    assert isinstance(_util.string_to_bool('1'), bool)
    assert _util.string_to_bool('1')


@pytest.mark.conversion
def test_string_to_bool_false():
    assert isinstance(_util.string_to_bool('0'), bool)
    assert not _util.string_to_bool('0')


@pytest.mark.conversion
def test_string_to_bool_invalid_value():
    with pytest.raises(ValueError) as exc_info:
        _util.string_to_bool('foo')
    assert str(exc_info.value) == 'Not a boolean value: {0!r}'.format('foo')


def test_udev_list_iterate_no_entry():
    assert not list(_util.udev_list_iterate(None))


def test_udev_list_iterate_mock():
    test_list = ['spam', 'eggs', 'foo', 'bar']
    get_list_entry = 'udev_enumerate_get_list_entry'
    with pytest.patch_libudev_list(test_list, get_list_entry) as get_list_entry:
        items = list(_util.udev_list_iterate(get_list_entry()))
        assert items == [
            ('spam', 'value of spam'),
            ('eggs', 'value of eggs'),
            ('foo', 'value of foo'),
            ('bar', 'value of bar')]


def raise_valueerror():
    raise ValueError('from function')

def test_reraise():
    with pytest.raises(ValueError) as excinfo:
        try:
            raise_valueerror()
        except ValueError as error:
            assert str(error) == 'from function'
            tb = sys.exc_info()[2]
            _util.reraise(ValueError('from except clause'), tb)
    assert str(excinfo.value) == 'from except clause'
    assert excinfo.traceback.getcrashentry().name == 'raise_valueerror'


def test_get_device_type_character_device():
    assert _util.get_device_type('/dev/console') == 'char'


def test_get_device_type_block_device():
    assert _util.get_device_type('/dev/sda') == 'block'


def test_get_device_type_no_device_file(tmpdir):
    filename = tmpdir.join('test')
    filename.ensure(file=True)
    with pytest.raises(ValueError) as excinfo:
        _util.get_device_type(str(filename))
    message = 'not a device file: {0!r}'.format(str(filename))
    assert str(excinfo.value) == message


def test_get_device_type_not_existing(tmpdir):
    filename = tmpdir.join('test')
    assert not tmpdir.check(file=True)
    with pytest.raises(EnvironmentError) as excinfo:
        _util.get_device_type(str(filename))
    assert excinfo.value.errno == errno.ENOENT
    assert excinfo.value.strerror == os.strerror(errno.ENOENT)
    assert excinfo.value.filename == str(filename)



# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import sys
import os

from docutils import nodes
from docutils.parsers.rst import Directive

# add the pyudev source directory to our path
doc_directory = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.normpath(
    os.path.join(doc_directory, os.pardir)))


class Mock(object):
    """
    Mock modules.

    Taken from
    http://read-the-docs.readthedocs.org/en/latest/faq.html#i-get-import-errors-on-libraries-that-depend-on-c-modules
    with some slight changes.
    """

    @classmethod
    def mock_modules(cls, *modules):
        for module in modules:
            sys.modules[module] = cls()

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return self.__class__()

    def __getattr__(self, attribute):
        if attribute in ('__file__', '__path__'):
            return os.devnull
        else:
            # return the *class* object here.  Mocked attributes may be used as
            # base class in pyudev code, thus the returned mock object must
            # behave as class, or else Sphinx autodoc will fail to recognize
            # the mocked base class as such, and "autoclass" will become
            # meaningless
            return self.__class__


# mock out native modules used throughout pyudev to enable Sphinx autodoc even
# if these modules are unavailable, as on readthedocs.org
Mock.mock_modules('PyQt4', 'PyQt4.QtCore', 'PySide', 'PySide.QtCore',
                  'glib', 'gobject', 'wx', 'wx.lib', 'wx.lib.newevent',
                  'pyudev._libudev')

# mock out the NewEvent function of wxPython.  Let's praise the silly wx API
def NewEventMock():
    yield 'event_class'
    yield 'event_constant'

sys.modules['wx.lib.newevent'].NewEvent = NewEventMock


import pyudev

needs_sphinx = '1.0'

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.intersphinx',
              'sphinxcontrib.issuetracker']

master_doc = 'index'
exclude_patterns = ['_build/*']
source_suffix = '.rst'

project = u'pyudev'
copyright = u'2010, 2011 Sebastian Wiesner'
version = '.'.join(pyudev.__version__.split('.')[:2])
release = pyudev.__version__

templates_path = ['_templates']
html_theme = 'default'
html_static_path = []
html_sidebars = {'**': ['info.html', 'localtoc.html', 'relations.html',
                        'sourcelink.html']}

intersphinx_mapping = {'python': ('http://docs.python.org/', None),
                       'pyside': ('http://www.pyside.org/docs/pyside/', None)}

issuetracker = 'github'
issuetracker_project = 'lunaryorn/pyudev'


class UDevVersion(Directive):
    """
    Directive to document the minimum udev version to use an attribute or
    method
    """
    has_content = False
    required_arguments = 1
    option_spec = {}

    def run(self):
        udevversion = self.arguments[0]
        para = nodes.paragraph(udevversion, '', classes=['udevversion'])
        text = 'Required udev version: {0}'.format(*self.arguments)
        para.append(nodes.inline(udevversion, text, classes=['versionmodified']))
        return [para]


def setup(app):
    from sphinx.ext.autodoc import cut_lines
    app.connect(b'autodoc-process-docstring', cut_lines(2, what=['module']))
    app.add_directive('udevversion', UDevVersion)



# -*- coding: utf-8 -*-
# Copyright (C) 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import re
from subprocess import check_call
from tempfile import NamedTemporaryFile
from xml.etree import ElementTree as etree

import pytest

from pyudev import _libudev as binding

libudev = binding.libudev


# new functions added in specific libudev versions
UDEV_ADDITIONS = {
    152: ['udev_new_from_environment'],
    154: ['udev_device_get_tags_list_entry',
          'udev_enumerate_add_match_tag',
          'udev_monitor_filter_add_match_tag'],
    165: ['udev_device_get_is_initialized',
          'udev_device_get_usec_since_initialized',
          'udev_enumerate_add_match_is_initialized'],
    167: ['udev_get_run_path',
          'udev_device_get_sysattr_list_entry'],
    172: ['udev_device_has_tag', 'udev_enumerate_add_match_parent'],
    }


LIBUDEV_H = '/usr/include/libudev.h'


def get_libudev_functions():
    with NamedTemporaryFile() as stream:
        check_call(['gccxml', LIBUDEV_H, '-fxml={0}'.format(stream.name)])
        tree = etree.parse(stream.name)
    names = [f.get('name') for f in tree.findall('.//Function')]
    return [n for n in names if n.startswith('udev_')]


def pytest_generate_tests(metafunc):
    if 'funcname' in metafunc.funcargnames:
        for namespace, members in binding.SIGNATURES.items():
            for funcname in members:
                full_name = '{0}_{1}'.format(namespace, funcname)
                metafunc.addcall(param=(namespace, funcname), id=full_name,
                                 funcargs=dict(funcname=full_name))
    if metafunc.function.__name__ == 'test_missing_functions':
        for version, functions in UDEV_ADDITIONS.items():
            for function in functions:
                funcargs = dict(version=version, missing_function=function)
                testid = 'version {0}, {1}'.format(version, function)
                metafunc.addcall(funcargs=funcargs, id=testid)
    elif metafunc.function.__name__ == 'test_is_wrapped':
        for function in get_libudev_functions():
            metafunc.addcall(funcargs=dict(function_name=function), id=function)


def pytest_funcarg__signature(request):
    namespace, name = request.param
    return binding.SIGNATURES[namespace][name]


def pytest_funcarg__argtypes(request):
    argtypes, _ = request.getfuncargvalue('signature')
    return argtypes


def pytest_funcarg__restype(request):
    _, restype = request.getfuncargvalue('signature')
    return restype


def pytest_funcarg__errcheck(request):
    funcname = request.getfuncargvalue('funcname')
    return binding.ERROR_CHECKERS.get(funcname)


@pytest.need_udev_version('>= {0}'.format(max(UDEV_ADDITIONS)))
def test_presence(funcname):
    assert hasattr(libudev, funcname)


def test_signatures(funcname, restype, argtypes, errcheck):
    func = getattr(libudev, funcname, None)
    if not func:
        pytest.skip('{0} not available'.format(funcname))
    assert func.restype == restype
    if not argtypes:
        assert not func.argtypes
    else:
        assert func.argtypes == argtypes
    assert func.errcheck == errcheck


def test_missing_functions(version, missing_function):
    if pytest.config.udev_version >= version:
        pytest.skip('{0} already exists in {1}'.format(
            missing_function, pytest.config.udev_version))
    assert not hasattr(libudev, missing_function)


WRAPPER_BLACKLIST_PATTERNS = [
     # vararg functions not supported by ctypes
    'udev_set_log_fn',
    # superfluous in Python, because arbitrary attributes can be attached to
    # objects anyways
    'udev_set_userdata', 'udev_get_userdata',
    # superfluous, because context is already available in .context
    'udev_enumerate_get_udev', 'udev_monitor_get_udev', 'udev_device_get_udev',
    # superfluous, because Python provides already tools to filter lists
    'udev_list_entry_get_by_name',
    # undocumented in libudev manual
    'udev_device_get_seqnum',
    # all queue functions (queue interface is not wrapped)
    re.compile('^udev_queue_.*')
]


def _is_blacklisted(function_name):
    for pattern in WRAPPER_BLACKLIST_PATTERNS:
        if pytest.is_unicode_string(pattern):
            if function_name == pattern:
                return True
        else:
            if pattern.match(function_name):
                return True
    else:
        return False


@pytest.mark.coverage
def test_is_wrapped(function_name):
    wrapped_functions = set('{0}_{1}'.format(ns, member)
                            for ns, members in binding.SIGNATURES.items()
                            for member in members)
    if _is_blacklisted(function_name):
        assert function_name not in wrapped_functions
    else:
        assert function_name in wrapped_functions



# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import re
import os
import stat
import operator
import sys
import gc
import errno
from itertools import count
from datetime import timedelta

import pytest
import mock

from pyudev import (Device,
                    DeviceNotFoundAtPathError,
                    DeviceNotFoundByNameError,
                    DeviceNotFoundByNumberError,
                    DeviceNotFoundInEnvironmentError)
from pyudev.device import Attributes, Tags


def pytest_generate_tests(metafunc):
    args = metafunc.funcargnames
    if any(a in ('sys_path', 'device_path', 'device') for a in args):
        devices = pytest.get_device_sample(metafunc.config)
        for device_path in devices:
            metafunc.addcall(id=device_path, param=device_path)
    elif 'operator' in args:
        for op in (operator.gt, operator.lt, operator.le, operator.ge):
            metafunc.addcall(funcargs=dict(operator=op), id=op.__name__)


class TestDevice(object):

    def test_from_path(self, context, device_path, sys_path):
        device = Device.from_path(context, device_path)
        assert device is not None
        assert device == Device.from_sys_path(context, sys_path)
        assert device == Device.from_path(context, sys_path)

    def test_from_path_strips_leading_slash(self, context):
        assert Device.from_path(context, 'devices/platform') == \
               Device.from_path(context, '/devices/platform')

    def test_from_sys_path(self, context, sys_path):
        device = Device.from_sys_path(context, sys_path)
        assert device is not None
        assert device.sys_path == sys_path

    def test_from_sys_path_device_not_found(self, context):
        sys_path = 'there_will_not_be_such_a_device'
        with pytest.raises(DeviceNotFoundAtPathError) as exc_info:
            Device.from_sys_path(context, sys_path)
        error = exc_info.value
        assert error.sys_path == sys_path
        assert str(error) == 'No device at {0!r}'.format(sys_path)

    def test_from_name(self, context, device):
        new_device = Device.from_name(context, device.subsystem,
                                      device.sys_name)
        assert new_device == device

    def test_from_name_no_device_in_existing_subsystem(self, context):
        with pytest.raises(DeviceNotFoundByNameError) as exc_info:
            Device.from_name(context, 'block', 'foobar')
        error = exc_info.value
        assert error.subsystem == 'block'
        assert error.sys_name == 'foobar'
        assert str(error) == 'No device {0!r} in {1!r}'.format(
            error.sys_name, error.subsystem)

    def test_from_name_nonexisting_subsystem(self, context):
        with pytest.raises(DeviceNotFoundByNameError) as exc_info:
            Device.from_name(context, 'no_such_subsystem', 'foobar')
        error = exc_info.value
        assert error.subsystem == 'no_such_subsystem'
        assert error.sys_name == 'foobar'
        assert str(error) == 'No device {0!r} in {1!r}'.format(
            error.sys_name, error.subsystem)

    def test_from_device_number(self, context, device_path, device_number,
                                device_node):
        if not device_node:
            pytest.skip('no device node, no device number')
        mode = os.stat(device_node).st_mode
        type = 'block' if stat.S_ISBLK(mode) else 'char'
        device = Device.from_device_number(context, type, device_number)
        assert device.device_number == device_number
        # make sure, we are really referring to the same device
        assert device.device_path == device_path

    def test_from_device_number_wrong_type(self, context, device_path,
                                           device_number, device_node):
        if not device_node:
            pytest.skip('no device node, no device number')
        mode = os.stat(device_node).st_mode
        # deliberately use the wrong type here to cause either failure or at
        # least device mismatch
        type = 'char' if stat.S_ISBLK(mode) else 'block'
        try:
            # this either fails, in which case the caught exception is raised,
            # or succeeds, but returns a wrong device (device numbers are not
            # unique across device types)
            device = Device.from_device_number(context, type, device_number)
            # if it succeeds, the resulting device must not match the one, we
            # are actually looking for!
            assert device.device_path != device_path
        except DeviceNotFoundByNumberError as error:
            # check the correctness of the exception attributes
            assert error.device_type == type
            assert error.device_number == device_number

    def test_from_device_number_invalid_type(self, context):
        with pytest.raises(ValueError) as exc_info:
            Device.from_device_number(context, 'foobar', 100)
        assert str(exc_info.value) == ('Invalid type: {0!r}. Must be one of '
                                       '"char" or "block".'.format('foobar'))

    def test_from_device_file(self, context, device_path, device_node):
        if not device_node:
            pytest.skip('no device file')
        device = Device.from_device_file(context, device_node)
        assert device.device_node == device_node
        assert device.device_path == device_path

    def test_from_device_file_links(self, context, device_links,
                                    device_node, device_path):
        if not device_links:
            pytest.skip('no device links')
        for link in device_links:
            link = os.path.join(context.device_path, link)
            device = Device.from_device_file(context, link)
            assert device.device_path == device_path
            assert link in device_links

    def test_from_device_file_no_device_file(self, context, tmpdir):
        filename = tmpdir.join('test')
        filename.ensure(file=True)
        with pytest.raises(ValueError) as excinfo:
            Device.from_device_file(context, str(filename))
        message = 'not a device file: {0!r}'.format(str(filename))
        assert str(excinfo.value) == message

    def test_from_device_file_non_existing(self, context, tmpdir):
        filename = tmpdir.join('test')
        assert not tmpdir.check(file=True)
        with pytest.raises(EnvironmentError) as excinfo:
            Device.from_device_file(context, str(filename))
        assert excinfo.value.errno == errno.ENOENT
        assert excinfo.value.strerror == os.strerror(errno.ENOENT)
        assert excinfo.value.filename == str(filename)

    @pytest.need_udev_version('>= 152')
    def test_from_environment(self, context):
        # there is no device in a standard environment
        with pytest.raises(DeviceNotFoundInEnvironmentError):
            Device.from_environment(context)

    def test_parent(self, device):
        assert device.parent is None or isinstance(device.parent, Device)

    @pytest.need_udev_version('>= 172')
    def test_child_of_parent(self, device):
        if device.parent is None:
            pytest.skip('Device {0!r} has no parent'.format(device))
        else:
            assert device in device.parent.children

    @pytest.need_udev_version('>= 172')
    def test_children(self, device):
        children = list(device.children)
        if not children:
            pytest.skip('Device {0!r} has no children'.format(device))
        else:
            assert all(c != device for c in children)
            assert all(c.parent == device for c in children)

    def test_find_parent(self, device):
        parent = device.find_parent(device.subsystem)
        if not parent:
            pytest.xfail('no parent within the same subsystem')
        assert parent.subsystem == device.subsystem
        assert parent in device.traverse()

    def test_find_parent_no_devtype_mock(self, device):
        get_parent = 'udev_device_get_parent_with_subsystem_devtype'
        ref = 'udev_device_ref'
        with pytest.nested(pytest.patch_libudev(get_parent),
                           pytest.patch_libudev(ref)) as (get_parent, ref):
            get_parent.return_value = mock.sentinel.device
            ref.return_value = mock.sentinel.referenced_device
            parent = device.find_parent('subsystem')
            get_parent.assert_called_with(device, b'subsystem', None)
            ref.assert_called_with(mock.sentinel.device)
            assert isinstance(get_parent.call_args[0][1], bytes)
            assert isinstance(parent, Device)
            assert parent._as_parameter_ is mock.sentinel.referenced_device

    def test_find_parent_with_devtype_mock(self, device):
        get_parent = 'udev_device_get_parent_with_subsystem_devtype'
        ref = 'udev_device_ref'
        with pytest.nested(pytest.patch_libudev(get_parent),
                           pytest.patch_libudev(ref)) as (get_parent, ref):
            get_parent.return_value = mock.sentinel.device
            ref.return_value = mock.sentinel.referenced_device
            parent = device.find_parent('subsystem', 'devtype')
            get_parent.assert_called_with(device, b'subsystem', b'devtype')
            ref.assert_called_with(mock.sentinel.device)
            args = get_parent.call_args[0][1:]
            assert all(isinstance(a, bytes) for a in args)
            assert isinstance(parent, Device)
            assert parent._as_parameter_ is mock.sentinel.referenced_device

    def test_traverse(self, device):
        child = device
        for parent in device.traverse():
            assert parent == child.parent
            child = parent

    def test_sys_path(self, device, sys_path):
        assert device.sys_path == sys_path
        assert pytest.is_unicode_string(device.sys_path)

    def test_device_path(self, device, device_path):
        assert device.device_path == device_path
        assert pytest.is_unicode_string(device.device_path)

    def test_subsystem(self, device, properties):
        assert device.subsystem == properties['SUBSYSTEM']
        assert pytest.is_unicode_string(device.subsystem)

    def test_device_sys_name(self, device):
        assert device.sys_name == os.path.basename(device.device_path)
        assert pytest.is_unicode_string(device.sys_name)

    def test_sys_number(self, device):
        match = re.search(r'\d+$', device.sys_name)
        # filter out devices with completely nummeric names (first character
        # doesn't count according to the implementation of libudev)
        if match and match.start() > 1:
            assert device.sys_number == match.group(0)
            assert pytest.is_unicode_string(device.sys_name)
        else:
            assert device.sys_number is None

    def test_type(self, device, properties):
        if 'DEVTYPE' in properties:
            assert device.device_type == properties['DEVTYPE']
            assert pytest.is_unicode_string(device.device_type)
        else:
            assert device.device_type is None

    def test_driver(self, device, properties):
        if 'DRIVER' in properties:
            assert device.driver == properties['DRIVER']
            assert pytest.is_unicode_string(device.driver)
        else:
            assert device.driver is None

    def test_device_node(self, device, device_node):
        if device_node:
            assert device.device_node == device_node
            assert pytest.is_unicode_string(device.device_node)
        else:
            assert device.device_node is None

    def test_device_number(self, device, device_number):
        assert device.device_number == device_number

    @pytest.need_udev_version('>= 165')
    def test_is_initialized(self, device):
        assert isinstance(device.is_initialized, bool)
        get_is_initialized = 'udev_device_get_is_initialized'
        with pytest.patch_libudev(get_is_initialized) as get_is_initialized:
            get_is_initialized.return_value = True
            assert device.is_initialized
            get_is_initialized.assert_called_with(device)

    @pytest.need_udev_version('>= 165')
    def test_time_since_initialized(self, device):
        assert isinstance(device.time_since_initialized, timedelta)
        usec_since_init = 'udev_device_get_usec_since_initialized'
        with pytest.patch_libudev(usec_since_init) as usec_since_init:
            usec_since_init.return_value = 100
            assert device.time_since_initialized.microseconds == 100
            usec_since_init.assert_called_with(device)

    def test_links(self, context, device, device_links):
        assert sorted(device.device_links) == sorted(
            os.path.join(context.device_path, l) for l in device_links)
        assert all(pytest.is_unicode_string(l) for l in device.device_links)

    def test_attributes(self, device):
        # see TestAttributes for complete attribute tests
        assert isinstance(device.attributes, Attributes)

    def test_no_leak(self, context):
        """
        Regression test for issue #32, modelled after the script which revealed
        this issue.

        The leak was caused by the following reference cycle between
        ``Attributes`` and ``Device``:

        Device._attributes -> Attributes.device

        https://github.com/lunaryorn/pyudev/issues/32
        """
        for _ in context.list_devices(subsystem='usb'):
            pass
        # make sure that no memory leaks
        assert not gc.garbage

    def test_tags(self, device):
        # see TestTags for complete tag tests
        assert isinstance(device.tags, Tags)

    def test_iteration(self, device, properties):
        device_properties = set(device)
        assert all(p in device_properties for p in properties)
        assert all(pytest.is_unicode_string(p) for p in device_properties)

    def test_length(self, device, all_properties):
        assert len(device) == len(all_properties)

    def test_getitem(self, device, properties):
        assert all(device[p] == properties[p] for p in properties)

    def test_getitem_devname(self, context, device, all_properties):
        if 'DEVNAME' not in all_properties:
            pytest.xfail('%r has no DEVNAME' % device)
        assert os.path.join(context.device_path, device['DEVNAME']) == \
               os.path.join(context.device_path, all_properties['DEVNAME'])

    def test_getitem_nonexisting(self, device):
        with pytest.raises(KeyError) as excinfo:
            device['a non-existing property']
        assert str(excinfo.value) == repr('a non-existing property')

    def test_asint(self, device, properties):
        for property in properties:
            value = properties[property]
            try:
                value = int(value)
            except ValueError:
                with pytest.raises(ValueError):
                    device.asint(property)
            else:
                assert device.asint(property) == value

    def test_asbool(self, device, properties):
        for property in properties:
            value = properties[property]
            if value == '1':
                assert device.asbool(property)
            elif value == '0':
                assert not device.asbool(property)
            else:
                with pytest.raises(ValueError) as exc_info:
                    device.asbool(property)
                message = 'Not a boolean value: {0!r}'
                assert str(exc_info.value) == message.format(value)

    def test_hash(self, device):
        assert hash(device) == hash(device.device_path)
        assert hash(device.parent) == hash(device.parent)
        assert hash(device.parent) != hash(device)

    def test_equality(self, device):
        assert device == device.device_path
        assert device == device
        assert device.parent == device.parent
        assert not (device == device.parent)

    def test_inequality(self, device):
        assert not (device != device.device_path)
        assert not (device != device)
        assert not (device.parent != device.parent)
        assert device != device.parent

    def test_device_ordering(self, platform_device, operator):
        with pytest.raises(TypeError) as exc_info:
            operator(platform_device, platform_device)
        assert str(exc_info.value) == 'Device not orderable'


class TestAttributes(object):

    def test_length(self, device):
        counter = count()
        for _ in device.attributes:
            next(counter)
        assert len(device.attributes) == next(counter)

    def test_iteration(self, device, attributes):
        device_attributes = set(device.attributes)
        assert all(a in device_attributes for a in attributes)
        assert all(pytest.is_unicode_string(a) for a in device_attributes)

    @pytest.need_udev_version('>= 167')
    def test_iteration_mock(self, device):
        attributes = [b'spam', b'eggs']
        get_sysattr_list = 'udev_device_get_sysattr_list_entry'
        with pytest.patch_libudev_list(attributes,
                                       get_sysattr_list) as get_sysattr_list:
            attrs = list(device.attributes)
            assert attrs == ['spam', 'eggs']
            get_sysattr_list.assert_called_with(device)

    def test_contains(self, device, attributes):
        assert all(a in device.attributes for a in attributes)

    def test_getitem(self, device, attributes):
        assert all(isinstance(device.attributes[a], bytes) for a in attributes)
        assert all(device.attributes[a] == v.encode(sys.getfilesystemencoding())
                   for a, v in attributes.items())

    def test_getitem_nonexisting(self, device):
        with pytest.raises(KeyError) as excinfo:
            device.attributes['a non-existing attribute']
        print(type(excinfo.value))
        assert str(excinfo.value) == repr('a non-existing attribute')

    def test_asstring(self, device, attributes):
        assert all(pytest.is_unicode_string(device.attributes.asstring(a))
                   for a in attributes)
        assert all(device.attributes.asstring(a) == v
                   for a, v in attributes.items())

    def test_asint(self, device, attributes):
        for attribute, value in attributes.items():
            try:
                value = int(value)
            except ValueError:
                with pytest.raises(ValueError):
                    device.attributes.asint(attribute)
            else:
                assert device.attributes.asint(attribute) == value

    def test_asbool(self, device, attributes):
        for attribute, value in attributes.items():
            if value == '1':
                assert device.attributes.asbool(attribute)
            elif value == '0':
                assert not device.attributes.asbool(attribute)
            else:
                with pytest.raises(ValueError) as exc_info:
                    device.attributes.asbool(attribute)
                message = 'Not a boolean value: {0!r}'
                assert str(exc_info.value) == message.format(value)


class TestTags(object):

    pytestmark = pytest.need_udev_version('>= 154')

    def test_iteration(self, device, tags):
        assert set(device.tags) == set(tags)
        assert all(pytest.is_unicode_string(t) for t in device.tags)

    def test_contains(self, device, tags):
        assert all(t in device.tags for t in tags)

    @pytest.need_udev_version('>= 172')
    def test_contains_mock(self, device):
        """
        Test that ``udev_device_has_tag`` is called if available.
        """
        has_tag = 'udev_device_has_tag'
        with pytest.patch_libudev(has_tag) as has_tag:
            has_tag.return_value = 1
            assert 'foo' in device.tags
            has_tag.assert_called_with(device, b'foo')



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


import sys
try:
    import setuptools
except ImportError:
    import distribute_setup
    distribute_setup.use_setuptools()
    import setuptools

from setuptools.command.build_py import build_py
import setuptools.command.build_py

import pyudev

def find_fixers(blacklist=None):
    blacklist = blacklist or []
    names = getattr(build_py, 'fixer_names', None) or []
    from lib2to3.refactor import get_fixers_from_package
    for p in setuptools.lib2to3_fixer_packages:
        names.extend(get_fixers_from_package(p))
    # explicitly remove all blacklisted fixers to trigger value errors on
    # non-existing filters in the blacklist
    for f in blacklist:
        names.remove(f)
    build_py.fixer_names = names

if sys.version_info[0] < 3:
    from codecs import open
    extra_arguments = {}
else:
    # Remove import fixer, because it blews up absolute imports on some python
    # versions, see Python bug #8358
    find_fixers(blacklist=['lib2to3.fixes.fix_import'])
    extra_arguments = dict(use_2to3=True)


with open('README.rst', encoding='utf-8') as stream:
    long_description = stream.read()


setuptools.setup(
    name='pyudev',
    version=str(pyudev.__version__),
    url='http://pyudev.readthedocs.org/',
    author='Sebastian Wiesner',
    author_email='lunaryorn@googlemail.com',
    description='A libudev binding',
    long_description=long_description,
    platforms=['Linux'],
    license='LGPL 2.1+',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Hardware',
        'Topic :: System :: Operating System Kernels :: Linux',
        ],
    packages=setuptools.find_packages(),
    **extra_arguments
    )



# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import sys
import socket
import errno
from contextlib import contextmanager
from select import select

import pytest
import mock

from pyudev import Monitor, MonitorObserver, Device

# many tests just consist of some monkey patching to test, that the Monitor
# class actually calls out to udev, correctly passing arguments and handling
# return value.  Actual udev calls are difficult to test, as return values
# and side effects are dynamic and environment-dependent.  It isn't
# necessary anyway, libudev can just assumed to be correct.


def _assert_from_netlink_called(context, *args):
    new_from_netlink = 'udev_monitor_new_from_netlink'
    with pytest.patch_libudev(new_from_netlink) as new_from_netlink:
        source = args[0].encode('ascii') if args else b'udev'
        new_from_netlink.return_value = mock.sentinel.pointer
        monitor = Monitor.from_netlink(context, *args)
        new_from_netlink.assert_called_with(context, source)
        assert isinstance(new_from_netlink.call_args[0][1], bytes)
        assert monitor._as_parameter_ is mock.sentinel.pointer


@contextmanager
def patch_filter_by(type):
    add_match = 'udev_monitor_filter_add_match_{0}'.format(type)
    filter_update = 'udev_monitor_filter_update'
    with pytest.patch_libudev(add_match) as add_match:
        add_match.return_value = 0
        with pytest.patch_libudev(filter_update) as filter_update:
            filter_update.return_value = 0
            yield add_match, filter_update


class TestMonitor(object):

    def test_from_netlink_invalid_source(self, context):
        with pytest.raises(ValueError) as exc_info:
            Monitor.from_netlink(context, source='invalid_source')
        message = ('Invalid source: {0!r}. Must be one of "udev" '
                   'or "kernel"'.format('invalid_source'))
        assert str(exc_info.value) == message

    def test_from_netlink_source_udev(self, context):
        monitor = Monitor.from_netlink(context)
        assert monitor._as_parameter_
        monitor = Monitor.from_netlink(context, source='udev')
        assert monitor._as_parameter_

    def test_from_netlink_source_udev_mock(self, context):
        _assert_from_netlink_called(context)
        _assert_from_netlink_called(context, 'udev')

    def test_from_netlink_source_kernel(self, context):
        monitor = Monitor.from_netlink(context, source='kernel')
        assert monitor._as_parameter_

    def test_from_netlink_source_kernel_mock(self, context):
        _assert_from_netlink_called(context, 'kernel')

    def test_from_socket(self, context, socket_path):
        monitor = Monitor.from_socket(context, str(socket_path))
        assert monitor._as_parameter_

    def test_from_socket_mock(self, context, socket_path):
        socket_path = str(socket_path)
        new_from_socket = 'udev_monitor_new_from_socket'
        with pytest.patch_libudev(new_from_socket) as new_from_socket:
            new_from_socket.return_value = mock.sentinel.pointer
            monitor = Monitor.from_socket(context, socket_path)
            new_from_socket.assert_called_with(
                context, socket_path.encode(sys.getfilesystemencoding()))
            assert monitor._as_parameter_ is mock.sentinel.pointer
            Monitor.from_socket(context, 'foobar')
            new_from_socket.assert_called_with(context, b'foobar')
            assert isinstance(new_from_socket.call_args[0][1], bytes)

    def test_fileno(self, monitor):
        # we can't do more than check that no exception is thrown
        monitor.fileno()

    def test_fileno_mock(self, monitor):
        get_fd = 'udev_monitor_get_fd'
        with pytest.patch_libudev(get_fd) as get_fd:
            get_fd.return_value = mock.sentinel.fileno
            assert monitor.fileno() is mock.sentinel.fileno
            get_fd.assert_called_with(monitor)

    def test_filter_by_no_subsystem(self, monitor):
        with pytest.raises(AttributeError):
            monitor.filter_by(None)

    def test_filter_by_subsystem_no_dev_type(self, monitor):
        monitor.filter_by(b'input')
        monitor.filter_by('input')

    def test_filter_by_subsystem_no_dev_type_mock(self, monitor):
        with patch_filter_by('subsystem_devtype') as (add_match, filter_update):
            monitor.filter_by(b'input')
            add_match.assert_called_with(monitor, b'input', None)
            filter_update.assert_called_with(monitor)
            monitor.filter_by('input')
            add_match.assert_called_with(monitor, b'input', None)
            filter_update.assert_called_with(monitor)
            assert isinstance(add_match.call_args[0][1], bytes)

    def test_filter_by_subsystem_dev_type(self, monitor):
        monitor.filter_by('input', b'usb_interface')
        monitor.filter_by('input', 'usb_interface')

    def test_filter_by_subsystem_dev_type_mock(self, monitor):
        with patch_filter_by('subsystem_devtype') as (add_match, filter_update):
            monitor.filter_by(b'input', b'usb_interface')
            add_match.assert_called_with(monitor, b'input', b'usb_interface')
            filter_update.assert_called_with(monitor)
            monitor.filter_by('input', 'usb_interface')
            add_match.assert_called_with(monitor, b'input', b'usb_interface')
            filter_update.assert_called_with(monitor)
            assert isinstance(add_match.call_args[0][2], bytes)

    @pytest.need_udev_version('>= 154')
    def test_filter_by_tag(self, monitor):
        monitor.filter_by_tag('spam')

    @pytest.need_udev_version('>= 154')
    def test_filter_by_tag_mock(self, monitor):
        with patch_filter_by('tag') as (match_tag, filter_update):
            match_tag.return_value = 0
            monitor.filter_by_tag(b'spam')
            match_tag.assert_called_with(monitor, b'spam')
            filter_update.assert_called_with(monitor)
            monitor.filter_by_tag('eggs')
            match_tag.assert_called_with(monitor, b'eggs')
            filter_update.assert_called_with(monitor)
            assert isinstance(match_tag.call_args[0][1], bytes)

    def test_remove_filter(self, monitor):
        """
        The underlying ``udev_monitor_filter_remove()`` is apparently broken.
        It always causes ``EINVAL`` from ``setsockopt()``.
        """
        with pytest.raises(ValueError):
            monitor.remove_filter()

    def test_remove_filter_mock(self, monitor):
        remove = 'udev_monitor_filter_remove'
        update = 'udev_monitor_filter_update'
        with pytest.patch_libudev(remove) as remove:
            remove.return_value = 0
            with pytest.patch_libudev(update) as update:
                update.return_value = 0
                monitor.remove_filter()
                update.assert_called_with(monitor)
                remove.assert_called_with(monitor)

    def test_enable_receiving_netlink_kernel_source(self, context):
        monitor = Monitor.from_netlink(context, source='kernel')
        monitor.enable_receiving()

    def test_enable_receiving_socket(self, context, socket_path):
        monitor = Monitor.from_socket(context, str(socket_path))
        monitor.enable_receiving()

    def test_enable_receiving_bound_socket(self, context, socket_path):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        # cause an error by binding the socket, thus testing error handling in
        # this method
        sock.bind(str(socket_path))
        monitor = Monitor.from_socket(context, str(socket_path))
        with pytest.raises(EnvironmentError) as exc_info:
            monitor.enable_receiving()
        pytest.assert_env_error(exc_info.value, errno.EADDRINUSE, str(socket_path))

    def test_enable_receiving_alias(self):
        assert Monitor.start == Monitor.enable_receiving

    def test_enable_receiving_mock(self, monitor):
        with pytest.patch_libudev('udev_monitor_enable_receiving') as func:
            func.return_value = 0
            monitor.enable_receiving()
            func.assert_called_with(monitor)

    def test_set_receive_buffer_size_mock(self, monitor):
        set_receive_buffer_size = 'udev_monitor_set_receive_buffer_size'
        with pytest.patch_libudev(set_receive_buffer_size) as func:
            func.return_value = 0
            monitor.set_receive_buffer_size(1000)
            func.assert_called_with(monitor, 1000)

    def test_set_receive_buffer_size_privilege_error(self, monitor,
                                                     socket_path):
        with pytest.raises(EnvironmentError) as exc_info:
            monitor.set_receive_buffer_size(1000)
        pytest.assert_env_error(exc_info.value, errno.EPERM)

    @pytest.mark.privileged
    def test_receive_device(self, monitor):
        # forcibly unload the dummy module to avoid hangs
        pytest.unload_dummy()
        monitor.filter_by('net')
        monitor.enable_receiving()
        # load the dummy device to trigger an add event
        pytest.load_dummy()
        select([monitor], [], [])
        action, device = monitor.receive_device()
        assert action == 'add'
        assert device.subsystem == 'net'
        assert device.device_path == '/devices/virtual/net/dummy0'
        # and unload again
        pytest.unload_dummy()
        action, device = monitor.receive_device()
        assert action == 'remove'
        assert device.subsystem == 'net'
        assert device.device_path == '/devices/virtual/net/dummy0'

    def test_receive_device_mock(self, monitor):
        receive_device = 'udev_monitor_receive_device'
        get_action = 'udev_device_get_action'
        with pytest.nested(pytest.patch_libudev(receive_device),
                           pytest.patch_libudev(get_action)) as (receive_device,
                                                                 get_action):
            receive_device.return_value = mock.sentinel.pointer
            get_action.return_value = b'action'
            action, device = monitor.receive_device()
            assert action == 'action'
            assert pytest.is_unicode_string(action)
            assert isinstance(device, Device)
            assert device.context is monitor.context
            assert device._as_parameter_ is mock.sentinel.pointer
            get_action.assert_called_with(mock.sentinel.pointer)
            receive_device.assert_called_with(monitor)

    @pytest.mark.privileged
    def test_iter(self, monitor):
        pytest.unload_dummy()
        monitor.filter_by('net')
        monitor.enable_receiving()
        pytest.load_dummy()
        iterator = iter(monitor)
        action, device = next(iterator)
        assert action == 'add'
        assert device.subsystem == 'net'
        assert device.device_path == '/devices/virtual/net/dummy0'
        pytest.unload_dummy()
        action, device = next(iterator)
        assert action == 'remove'
        assert device.subsystem == 'net'
        assert device.device_path == '/devices/virtual/net/dummy0'
        iterator.close()


class TestMonitorObserver(object):

    def receive_event(self, action, device):
        self.events.append((action, device))
        if len(self.events) >= 2:
            self.observer.send_stop()

    def make_observer(self, monitor):
        self.observer = MonitorObserver(monitor, self.receive_event)
        return self.observer

    def setup(self):
        self.events = []

    def teardown(self):
        self.events = None

    def test_fake(self, fake_monitor, platform_device):
        observer = self.make_observer(fake_monitor)
        observer.start()
        fake_monitor.trigger_action('add')
        fake_monitor.trigger_action('remove')
        # fake one second for the tests to finish
        observer.join(1)
        # forcibly quit the thread if it is still alive
        if observer.is_alive():
            observer.stop()
        # check that we got two events
        assert self.events == [('add', platform_device), ('remove', platform_device)]

    @pytest.mark.privileged
    def test_real(self, context, monitor):
        observer = self.make_observer(monitor)
        pytest.unload_dummy()
        monitor.filter_by('net')
        monitor.enable_receiving()
        observer.start()
        pytest.load_dummy()
        pytest.unload_dummy()
        observer.join(2)
        if observer.is_alive():
            observer.stop()
        assert [e[0] for e in self.events] == ['add', 'remove']
        assert all(device.device_path == '/devices/virtual/net/dummy0' for _, device in self.events)




# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


"""
    pyudev._util
    ============

    Internal utilities

    .. moduleauthor::  Sebastian Wiesner  <lunaryorn@googlemail.com>
"""


from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os
import sys
import stat

from pyudev._libudev import libudev

if sys.version_info[0] == 2:
    from pyudev._py2util import *
else:
    from pyudev._py3util import *


def ensure_byte_string(value):
    """
    Return the given ``value`` as bytestring.

    If the given ``value`` is not a byte string, but a real unicode string, it
    is encoded with the filesystem encoding (as in
    :func:`sys.getfilesystemencoding()`).
    """
    if not isinstance(value, bytes):
        value = value.encode(sys.getfilesystemencoding())
    return value


def ensure_unicode_string(value):
    """
    Return the given ``value`` as unicode string.

    If the given ``value`` is not a unicode string, but a byte string, it is
    decoded with the filesystem encoding (as in
    :func:`sys.getfilesystemencoding()`).
    """
    if not isinstance(value, unicode):
        value = value.decode(sys.getfilesystemencoding())
    return value


def property_value_to_bytes(value):
    """
    Return a byte string, which represents the given ``value`` in a way
    suitable as raw value of an udev property.

    If ``value`` is a boolean object, it is converted to ``'1'`` or ``'0'``,
    depending on whether ``value`` is ``True`` or ``False``.  If ``value`` is a
    byte string already, it is returned unchanged.  Anything else is simply
    converted to a unicode string, and then passed to
    :func:`ensure_byte_string`.
    """
    # udev represents boolean values as 1 or 0, therefore an explicit
    # conversion to int is required for boolean values
    if isinstance(value, bool):
        value = int(value)
    if isinstance(value, bytes):
        return value
    else:
        return ensure_byte_string(unicode(value))


def string_to_bool(value):
    """
    Convert the given unicode string ``value`` to a boolean object.

    If ``value`` is ``'1'``, ``True`` is returned.  If ``value`` is ``'0'``,
    ``False`` is returned.  Any other value raises a
    :exc:`~exceptions.ValueError`.
    """
    if value not in ('1', '0'):
        raise ValueError('Not a boolean value: {0!r}'.format(value))
    return value == '1'


def udev_list_iterate(entry):
    """
    Iteration helper for udev list entry objects.

    Yield a tuple ``(name, value)``.  ``name`` and ``value`` are bytestrings
    containing the name and the value of the list entry.  The exact contents
    depend on the list iterated over.
    """
    while entry:
        name = libudev.udev_list_entry_get_name(entry)
        value = libudev.udev_list_entry_get_value(entry)
        yield (name, value)
        entry = libudev.udev_list_entry_get_next(entry)


# for the sake of readability
_is_char_device = stat.S_ISCHR
_is_block_device = stat.S_ISBLK


def get_device_type(filename):
    """
    Get the device type of a device file.

    ``filename`` is a string containing the path of a device file.

    Return ``'char'`` if ``filename`` is a character device, or ``'block'`` if
    ``filename`` is a block device.  Raise :exc:`~exceptions.ValueError` if
    ``filename`` is no device file at all.  Raise
    :exc:`~exceptions.EnvironmentError` if ``filename`` does not exist or if
    its metadata was inaccessible.

    .. versionadded:: 0.15
    """
    mode = os.stat(filename).st_mode
    if _is_char_device(mode):
        return 'char'
    elif _is_block_device(mode):
        return 'block'
    else:
        raise ValueError('not a device file: {0!r}'.format(filename))



# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

from itertools import count
from functools import partial

import pytest
from mock import Mock

class BaseBinding(object):
    def trigger_observer(self, action, monitor, action_trigger):
        mainloop = self.create_mainloop()

        signal_counter = count(1)

        def _quit_when_done(*args, **kwargs):
            value = next(signal_counter)
            if value >= 2:
                self.quit_mainloop(mainloop)

        # slot dummies
        event_slot = Mock(side_effect=_quit_when_done)
        action_slots = {action: Mock(side_effect=_quit_when_done)}
        observer = self.create_observer(monitor)
        self.connect_signals(observer, event_slot, action_slots)

        self.single_shot(0, action_trigger)
        self.single_shot(5000, partial(self.quit_mainloop, mainloop))
        self.run_mainloop(mainloop)
        self.cleanup(observer)
        return event_slot, action_slots


class Qt4Binding(BaseBinding):

    def __init__(self, bindingname):
        self.name = bindingname

    def create_observer(self, monitor):
        name = self.name.lower()
        mod = __import__('pyudev.{0}'.format(name), None, None, [name])
        observer = mod.QUDevMonitorObserver(monitor)
        assert observer.monitor is monitor
        assert isinstance(observer.notifier, self.qtcore.QSocketNotifier)
        return observer

    def import_or_skip(self):
        self.qtcore = pytest.importorskip('{0}.QtCore'.format(self.name))

    def create_mainloop(self):
        app = self.qtcore.QCoreApplication.instance()
        if not app:
            app = self.qtcore.QCoreApplication([])
        return app

    def quit_mainloop(self, app):
        app.quit()

    def run_mainloop(self, app):
        app.exec_()

    def connect_signals(self, observer, event_slot, action_slots):
        observer.deviceEvent.connect(event_slot)
        signal_map = {'add': observer.deviceAdded,
                      'remove': observer.deviceRemoved,
                      'change': observer.deviceChanged,
                      'move': observer.deviceMoved}
        for action, slot in action_slots.items():
            signal_map[action].connect(slot)

    def single_shot(self, timeout, callback):
        self.qtcore.QTimer.singleShot(timeout, callback)

    def cleanup(self, observer):
        observer.enabled = False


class GlibBinding(BaseBinding):
    name = 'Glib'

    def __init__(self):
        self.event_sources = []

    def create_observer(self, monitor):
        from pyudev.glib import GUDevMonitorObserver
        observer = GUDevMonitorObserver(monitor)
        assert observer.monitor is monitor
        assert isinstance(observer.event_source, int)
        return observer

    def import_or_skip(self):
        self.glib = pytest.importorskip('glib')
        pytest.importorskip('gobject')

    def create_mainloop(self):
        return self.glib.MainLoop()

    def quit_mainloop(self, mainloop):
        mainloop.quit()

    def run_mainloop(self, mainloop):
        mainloop.run()

    def wrap_callback(self, callback):
        """
        Remove the emitter from callback invocation, it is not of any
        interest to the test code.
        """
        def _wrapper(obj, *args, **kwargs):
            return callback(*args, **kwargs)
        return _wrapper

    def connect_signals(self, observer, event_slot, action_slots):
        observer.connect('device-event', self.wrap_callback(event_slot))
        signal_map = {'add': 'device-added', 'remove': 'device-removed',
                      'change': 'device-changed', 'move': 'device-moved'}
        for action, slot in action_slots.items():
            observer.connect(signal_map[action], self.wrap_callback(slot))

    def single_shot(self, timeout, callback):
        def _callback():
            callback()
            return False
        self.event_sources.append(self.glib.timeout_add(timeout, _callback))

    def cleanup(self, observer):
        observer.enabled = False
        # cleanup all timer event sources.  If this was not done, all tests
        # save the first would spin endlessly.  Apparently older event sources
        # kind of "confuse" the main loop of the current test
        for source in self.event_sources:
            self.glib.source_remove(source)
        self.event_sources = []


class WXBinding(BaseBinding):
    name = 'wx'

    timer = []

    def create_observer(self, monitor):
        from pyudev.wx import WxUDevMonitorObserver
        observer = WxUDevMonitorObserver(monitor)
        assert observer.monitor is monitor
        return observer

    def import_or_skip(self):
        self.wx = pytest.importorskip('wx')

    def create_mainloop(self):
        app = self.wx.App(False)
        app.frame = self.wx.Frame(None)
        self.frame = app.frame
        return app

    def quit_mainloop(self, app):
        app.frame.Close()

    def run_mainloop(self, app):
        app.MainLoop()

    def connect_signals(self, observer, event_slot, action_slots):
        def wrap_callback(callback):
            def _wrapper(event):
                if (hasattr(event, 'action')):
                    return callback(event.action, event.device)
                else:
                    return callback(event.device)
            return _wrapper

        import pyudev.wx
        observer.Bind(pyudev.wx.EVT_DEVICE_EVENT, wrap_callback(event_slot))
        event_map = {
                'add': pyudev.wx.EVT_DEVICE_ADDED,
                'remove': pyudev.wx.EVT_DEVICE_REMOVED,
                'change': pyudev.wx.EVT_DEVICE_CHANGED,
                'move': pyudev.wx.EVT_DEVICE_MOVED }
        for action, slot in action_slots.items():
            observer.Bind(event_map[action], wrap_callback(slot))

    def single_shot(self, timeout, callback):
        timer = self.wx.PyTimer(callback)
        self.timer.append(timer)
        timer.Start(timeout, True)

    def cleanup(self, observer):
        # disable the observer to avoid that some pending observers confuse our
        # tests
        observer.enabled = False


BINDINGS = [Qt4Binding('PyQt4'), Qt4Binding('PySide'), GlibBinding(), WXBinding()]
ACTIONS = ('add', 'remove', 'change', 'move')


def pytest_generate_tests(metafunc):
    if 'binding' in metafunc.funcargnames:
        for binding in BINDINGS:
            funcargs = dict(binding=binding)
            id = binding.name
            if 'action' in metafunc.funcargnames:
                for action in ACTIONS:
                    funcargs.update(action=action)
                    metafunc.addcall(funcargs=funcargs,
                                     id=id + ',' + action)
            else:
                metafunc.addcall(funcargs=funcargs, id=id)


def test_fake_monitor(fake_monitor, platform_device):
    """
    Test the fake monitor just to make sure, that it works.
    """
    for action in ('add', 'remove'):
        fake_monitor.trigger_action(action)
        received_action, device = fake_monitor.receive_device()
        assert action == received_action
        assert device == platform_device


class TestObserver(object):

    def test_observer_fake(self, binding, action, fake_monitor,
                           platform_device):
        binding.import_or_skip()
        event_slot, action_slots = binding.trigger_observer(
            action, fake_monitor, lambda: fake_monitor.trigger_action(action))
        # check, that both slots were called
        event_slot.assert_called_with(action, platform_device)
        action_slots.pop(action).assert_called_with(platform_device)

    @pytest.mark.privileged
    def test_observer(self, binding, monitor):
        binding.import_or_skip()
        pytest.unload_dummy()
        monitor.filter_by('net')
        monitor.enable_receiving()
        event_slot, action_slot = binding.trigger_observer(
            'add', monitor, pytest.load_dummy)
        action, device = event_slot.call_args[0]
        assert action == 'add'
        assert device.subsystem == 'net'
        assert device.device_path == '/devices/virtual/net/dummy0'
        event_slot, action_slot = binding.trigger_observer(
            'remove', monitor, pytest.unload_dummy)
        action, device = event_slot.call_args[0]
        assert action == 'remove'
        assert device.subsystem == 'net'
        assert device.device_path == '/devices/virtual/net/dummy0'
